package jess;

import java.io.PrintWriter;
import java.io.Serializable;
import java.util.*;

/**
 * A Defrule is a specific action meant to be taken when certain
 * conditions are met in working memory. Although you can construct
 * Defrules from the Java API, they are almost always parsed from
 * "defrule" constructs in the Jess language. Building a Defrule from
 * Java is very complicated.
 * <p/>
 * (C) 2006 Sandia National Laboratories<br>
 */

public class Defrule extends HasLHS implements Serializable, MatchInfoSource {
    private Funcall[] m_actions;
    private int m_nActions;
    private Value m_salienceVal;
    private boolean m_autoFocus = false;
    private boolean m_noLoop;
    private Set m_testedSlots = new HashSet();
    private String m_actKey;
    private String m_salKey;

    public Defrule(String name, String docstring, Rete engine) throws JessException {
        super(name, docstring, engine);
        // ###
        m_salienceVal = new Value(0, RU.INTEGER);
        m_actKey = engine.getNextNodeKey() + "A";
        m_salKey = engine.getNextNodeKey() + "S";
        // addJessListener(new PrintingListener());
    }

    /**
     * Fetch the current salience value for this rule. This method won't re-evaluate the salience;
     * the last value previously computed will be returned.
     *
     * @return the salience of this defrule
     */

    public final int getSalience(Rete engine) {
        return getSalienceHolder(engine)[0];
    }

    int[] getSalienceHolder(Rete engine) {
        int[] holder = ((int[]) engine.getKeyedStorage(m_salKey));
        if (holder == null) {
            holder = new int[1];
            engine.putKeyedStorage(m_salKey, holder);
        }
        return holder;
    }

    /**
     * Fetch the raw salience for this rule as a jess.Value. For rules with dynamic salience, the
     * Value returned will be a jess.FuncallValue.
     *
     * @return the salience of this defrule
     */

    public Value getSalienceValue() {
        return m_salienceVal;
    }

    /**
     * Set the salience of this rule. The value van be a fixed integer, or it can be a function call. The default salience is 0.
     *
     * @param v      the salience expression
     * @param engine a Rete object to provide an execution context
     * @throws JessException if anything goes wrong
     */
    public void setSalience(Value v, Rete engine) throws JessException {
        try {
            m_salienceVal = v;
            evalSalience(engine);
        } finally {
            if (getNext() != null)
                ((Defrule) getNext()).setSalience(v, engine);
        }
    }

    /**
     * Evaluate the salience of this rule. If the salience was set to a
     * Funcall value during parsing, then this function may return
     * different values over time. If the salience is constant, this is
     * equivalent to getSalience. The computed value will be the next value returned by
     * getSalience().
     *
     * @param engine The Rete engine the rule belongs to
     * @return The evaluated salience
     * @throws JessException If something goes wrong
     */
    public int evalSalience(Rete engine) throws JessException {
        int[] holder = getSalienceHolder(engine);
        Context gc = engine.getGlobalContext().push();
        try {
            holder[0] = m_salienceVal.intValue(gc);
        } finally {
            gc.pop();
        }
        return holder[0];
    }

    /**
     * Indicate whether this rule's salience is different from the default value.
     *
     * @return true if this rule's salience is not the default.
     */
    public boolean hasNonDefaultSalience() {
        try {
            return m_salienceVal.type() != RU.INTEGER || m_salienceVal.intValue(null) != 0;
        } catch (JessException e) {
            return true;
        }
    }

    /**
     * Indicate whether this rule will automatically focus its module when it is activated.
     *
     * @return the value of this property
     */
    public boolean getAutoFocus() {
        return m_autoFocus;
    }

    /**
     * Tell this rule whether to automatically focus the module it appears in when the rule is activated.
     * The default value is false.
     *
     * @param autoFocus the desired value of this property.
     */
    public void setAutoFocus(boolean autoFocus) {
        m_autoFocus = autoFocus;
        if (getNext() != null)
            ((Defrule) getNext()).setAutoFocus(autoFocus);
    }

    /**
     * Add a conditional element (a pattern or group of patterns) to this rule. Normally this will only
     * be called by the Jess parser or the XML parser, but it is theoretically possible for users to build their
     * own rules this way.
     *
     * @param ce     the conditional element to add
     * @param engine the rule engine that provides the execution context
     * @throws JessException if anything goes wrong
     */

    public void addCE(ConditionalElementX ce, Rete engine) throws JessException {
        if (ce.getLogical() && getGroupSize() > 0)
            if (!getLHSComponent(getGroupSize() - 1).getLogical())
                throw new JessException("Defrule.addCE",
                                        "Logical CEs can't follow non-logical CEs",
                                        m_name);

        int oldCount = getPatternCount();
        super.addCE(ce, engine);

        // Make a note of slots that should matter for per-slot activation
        ce.recordTestedSlots(oldCount, m_testedSlots);

    }

    boolean isSlotTested(int factIndex, int slotIndex) {
        TestedSlot ts = new TestedSlot(factIndex, slotIndex);
        return m_testedSlots.contains(ts);
    }

    boolean isSlotTested(Token token, Context context) {
        Fact fact = context.getSlotSpecificModifiedFact();
        String[] slotNames = context.getModifiedSlots();
        for (int i = 0; i < slotNames.length; i++) {
            String slotName = slotNames[i];
            int slotIndex = fact.getDeftemplate().getSlotIndex(slotName);
            int size = token.size();
            for (int factIndex = 0; factIndex < size; ++factIndex) {
                if (fact.equals(token.fact(factIndex).getIcon())) {
                    if (isSlotTested(factIndex, slotIndex))
                        return true;
                }
            }
        }
        return false;
    }

    boolean isSlotTested(int factIndex, Context context) {
        Fact fact = context.getSlotSpecificModifiedFact();
        String[] slotNames = context.getModifiedSlots();
        for (int i = 0; i < slotNames.length; i++) {
            String slotName = slotNames[i];
            int slotIndex = fact.getDeftemplate().getSlotIndex(slotName);
            if (isSlotTested(factIndex, slotIndex))
                return true;
        }
        return false;
    }

    private void doAddCall(int tag, Token token, Context context, Rete engine) throws JessException {
        if (shouldAddCall(tag, token, context)) {
            Activation a = new Activation(engine, token, this);
            engine.addActivation(a);
            getActivations(engine).put(token, a);
        }
    }

    private Map getActivations(Rete engine) {
        Map activations = (Map) engine.getKeyedStorage(m_actKey);
        if (activations == null) {
            activations = new HashMap();
            engine.putKeyedStorage(m_actKey, activations);
        }
        return activations;
    }

    private void removeCall(int tag, Token token, Context context) throws JessException {
        if (shouldRemoveCall(tag, token, context)) {
            Rete engine = context.getEngine();
            Activation a = (Activation) getActivations(engine).remove(token);
            if (a != null) {
                engine.removeActivation(a);
            }
        }
    }

    private boolean shouldRemoveCall(int tag, Token token, Context context) {
        if (tag == RU.MODIFY_REMOVE)
            return context.getSlotSpecificModifiedFact() == null || isSlotTested(token, context);
        else
            return true;
    }

    private boolean shouldAddCall(int tag, Token token, Context context) {
        if (m_noLoop) {
            Activation a = context.getEngine().getThisActivation();
            if (a != null && a.getRule() == this) {
                return false;
            }
        }

        if (tag == RU.MODIFY_ADD) {
            return context.getSlotSpecificModifiedFact() == null || isSlotTested(token, context);
        } else
            return true;
    }

    private void possiblyDoAddCall(Token token, Context context, Rete engine)
            throws JessException {
        // We're not new, so updates don't affect us
        if (!m_new)
            return;

        // We've already got this one
        if (getActivations(engine).get(token) != null)
            return;

        // Add a new activation
        doAddCall(RU.UPDATE, token, context, engine);
    }

    /**
     * An implementation detail, public only because Jave requires methods that implement an
     * interface to be public.
     */
    public void callNodeLeft(int tag, Token token, Context context)
            throws JessException {
        broadcastEvent(tag, JessEvent.RETE_TOKEN_LEFT, token);

        Rete engine = context.getEngine();
        switch (tag) {

            case RU.ADD:
            case RU.MODIFY_ADD:
                doAddCall(tag, token, context, engine);
                break;

            case RU.REMOVE:
            case RU.MODIFY_REMOVE:
                removeCall(tag, token, context);
                break;


            case RU.UPDATE:
                possiblyDoAddCall(token, context, engine);
                break;

            case RU.CLEAR:
                getActivations(engine).clear();
                break;
        }
    }

    void ready(Token fact_input, Context c) {

        super.ready(fact_input, c);

        // If our assertions will have logical support, put the needed
        // info into the context. We peel off the post-logical part of
        // the token that activated us and save only the "logical"
        // part. The assertFact() function will save information about
        // this token and any asserted facts into the Node2. The
        // Node2, in turn, will send messages to the Rete object when
        // that token is removed.

        if (m_logicalNode != null) {

            Token logicalToken = fact_input;
            int count = fact_input.size();

            int logicalSize = m_logicalNode.getTokenSize();
            while (count > logicalSize) {
                logicalToken = logicalToken.getParent();
                --count;
            }

            c.setLogicalSupportNode(m_logicalNode);
            c.setToken(logicalToken);
        }

        return;
    }


    /**
     * Do the RHS of this rule. For each action (ValueVector form of a
     * Funcall), do two things: 1) Call ExpandAction to do variable
     * substitution and subexpression expansion 2) call
     * Funcall.Execute on it.
     * <p/>
     * Fact_input is the Vector of ValueVector facts we were fired with.
     */
    void fire(Token factInput, Rete engine, Context context) throws JessException {
        getActivations(engine).remove(factInput);
        Context c = context.push();
        c.clearReturnValue();

        // Pull needed values out of facts into bindings table
        ready(factInput, c);


        try {
            // OK, now run the rule. For every action...
            for (int i = 0; i < m_nActions; i++) {
                m_actions[i].execute(c);

                if (c.returning()) {
                    c.clearReturnValue();
                    c.pop();
                    engine.popFocus(getModule());
                    return;
                }
            }
        } catch (BreakException bk) {
            // Fall through
        } catch (JessException re) {
            re.addContext("defrule " + getDisplayName());
            throw re;
        } finally {
            c.pop();
        }

    }

    void debugPrint(Token facts, int seq, PrintWriter ps) {
        ps.print("FIRE ");
        ps.print(seq);
        ps.print(" ");
        ps.print(getDisplayName());
        for (int i = 0; i < facts.size(); i++) {
            Fact f = facts.fact(i);
            if (f.getFactId() != -1)
                ps.print(" f-" + f.getFactId());
            if (i < facts.size() - 1)
                ps.print(",");
        }
        ps.println();
        ps.flush();
    }

    /**
     * Fetch the number of actions on this rule's RHS
     *
     * @return The number of actions
     */
    public int getNActions() {
        return m_nActions;
    }

    /**
     * Fetch the idx-th RHS action of this rule.
     *
     * @param idx The zero-based index of the action to fetch
     * @return the action as a function call
     */
    public Funcall getAction(int idx) {
        return m_actions[idx];
    }


    /**
     * Add an action to this defrule
     *
     * @param fc the action as a function call
     */
    public void addAction(Funcall fc) {
        if (m_actions == null || m_nActions == m_actions.length) {
            Funcall[] temp = m_actions;
            m_actions = new Funcall[m_nActions + 5];
            if (temp != null)
                System.arraycopy(temp, 0, m_actions, 0, m_nActions);
        }
        m_actions[m_nActions++] = fc;

        if (getNext() != null)
            ((Defrule) getNext()).addAction(fc);
    }

    /**
     * Defrule participates in the Visitor pattern. Will call accept(this) on the Visitor.
     *
     * @param jv a Visitor
     * @return whatever jv.accept(this) returns
     */
    public Object accept(Visitor jv) {
        return jv.visitDefrule(this);
    }

    private LogicalNode m_logicalNode = null;

    void setLogicalInformation(LogicalNode node) {
        m_logicalNode = node;
        node.setMatchInfoSource(this);
    }

    /**
     * Returns a brief description of this rule
     *
     * @return the String "Defrule rule-name"
     */
    public String toString() {
        return "Defrule " + getName();
    }

    // For testing
    LogicalNode getLogicalNode() {
        return m_logicalNode;
    }

    /**
     * Specify whether this rule can activate itself through actions it takes on its RHS. If this property is
     * set to true, then modifying a fact on the RHS of this rule can't reactivate the same rule, even if that
     * fact is matched by this rule. The default value is false.
     *
     * @param b the desired value of the property.
     */
    public void setNoLoop(boolean b) {
        m_noLoop = b;
        if (getNext() != null)
            ((Defrule) getNext()).setNoLoop(b);
    }

    /**
     * Indicate whether this rule can activate itself through actions it takes on its RHS. If this property is
     * true, then modifying a fact on the RHS of this rule can't reactivate the same rule, even if that
     * fact is matched by this rule.
     *
     * @return the current value of the property.
     */

    public boolean isNoLoop() {
        return m_noLoop;
    }

    void createChain(ConditionalElementX ce, Rete engine) throws JessException {
        doSetLHS(ce.getConditionalElementX(0), engine);
        Defrule rule = this;
        String name = getName();
        for (int i = 1; i < ce.getGroupSize(); ++i) {
            ConditionalElementX newPatterns = ce.getConditionalElementX(i);
            Defrule next = new Defrule(name + "&" + i, getDocstring(), engine);
            next.doSetLHS(newPatterns, engine);
            rule.setNext(next);
            rule = next;
        }
    }

    /**
     * Return the type of this construct
     *
     * @return the String "defrule"
     */
    public final String getConstructType() {
        return "defrule";
    }

    /**
     * Implementation detail. Public only because Java requires methods that implement an interface to
     * be public.
     */
    public boolean isRelevantChange(int factIndex, Token token, Context context) {
        Fact slotSpecificModifiedFact = context.getSlotSpecificModifiedFact();
        if (slotSpecificModifiedFact == null)
            return true;
        Fact tokenFact = token.fact(factIndex).getIcon();

        // TODO.  This isn't perfect: there could be some false positives here
        // for rules with multiple logical slot-specific "not"s. Sigh.
        if (tokenFact == Fact.getNullFact())
            return true;

        else if (!slotSpecificModifiedFact.equals(tokenFact))
            return false;

        return isSlotTested(factIndex, context);
    }
}



