package jess;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

/**
 <p>A Deftemplate defines a class of Jess facts. Its definition
 includes a list of properties or "slots", each of which can have a
 default value and a type. Deftemplates can be created by hand, but
 they usually come from executing the "deftemplate" construct or the
 "defclass" function in the Jess language.
 To build a useful template, you need to specify the name to the constructor,
 then add the named slots one by one:</p>

 <pre>
 Rete r = new Rete();
 Deftemplate dt = new Deftemplate("point", "A 2D point", r);
 Value zero = new Value(0, RU.INTEGER);
 dt.addSlot("x", zero, "NUMBER");
 dt.addSlot("y", zero, "NUMBER");
 r.addDeftemplate(dt);
 </pre>

 (C) 2006 Sandia National Laboratories<br>

 */
public class Deftemplate implements Serializable, Visitable, Modular {

    // Keys are type names, values are type codes
    private static final Map s_typeCodes = new HashMap();

    // Keys are type codes, values are type names
    private static final Map s_typeNames = new HashMap();

    static final String[] TYPE_NAMES = {
        "ANY",
        "INTEGER",
        "FLOAT",
        "NUMBER",
        "ATOM",
        "SYMBOL",
        "STRING",
        "LEXEME",
        "OBJECT",
        "LONG"
    };

    static final int[] TYPE_CODES = {
        RU.ANY,
        RU.INTEGER,
        RU.FLOAT,
        RU.NUMBER,
        RU.SYMBOL,
        RU.SYMBOL,
        RU.STRING,
        RU.LEXEME,
        RU.JAVA_OBJECT,
        RU.LONG
    };

    private static final Deftemplate
            s_rootTemplate = new Deftemplate(RU.ROOT_DEFTEMPLATE, "Parent template");
    private static final Deftemplate s_clearTemplate = new Deftemplate("__clear", "(Implied)");
    private static final Deftemplate s_nullTemplate = new Deftemplate("__not_or_test_CE", "(Implied)");
    private static final Deftemplate s_initialTemplate = new Deftemplate("initial-fact", "(Implied)");
    private static final Deftemplate s_testTemplate = new Deftemplate("test", "(Implied)");
    private static final Deftemplate s_accumulateTemplate = new Deftemplate(Accumulate.RESULT, "(Accumulate result)");

    private boolean m_slotSpecific;

    static {
        try {
            for (int i = 0; i < TYPE_NAMES.length; i++) {
                Value value = new Value(TYPE_CODES[i], RU.INTEGER);
                s_typeCodes.put(TYPE_NAMES[i], value);
                s_typeNames.put(value, TYPE_NAMES[i]);
            }

            s_accumulateTemplate.addSlot("value", Funcall.NIL, "ANY");
            s_testTemplate.addSlot("__data", Funcall.NILLIST, "ANY");

        } catch (JessException re) { /* can't happen */
        }
    }

    /**
     * The root template that serves as the ultimate parent of all
     * other templates. This template is to deftemplates as java.lang.Object is to
     * Java objects.
     * @return The singleton root template.
     */
    public static Deftemplate getRootTemplate() {
        return s_rootTemplate;
    }

    /**
     * The template for "initial-fact".
     * @return A deftemplate.
     */
    public static Deftemplate getInitialTemplate() {
        return s_initialTemplate;
    }

    /**
     * The template for a special fact used internally by Jess.
     * @return A deftemplate
     */
    public static Deftemplate getClearTemplate() {
        return s_clearTemplate;
    }

    /** The template for a special fact used internally by Jess.
     * @return A deftemplate
     */
    public static Deftemplate getNullTemplate() {
        return s_nullTemplate;
    }

    /** The template for a special fact used internally by Jess.
     * @return A deftemplate
     */
    public static Deftemplate getTestTemplate() {
        return s_testTemplate;
    }

    /** The template for a special fact used internally by Jess.
     * @return A deftemplate
     */
    public static Deftemplate getAccumTemplate() {
        return s_accumulateTemplate;
    }

    static void addStandardTemplates(Rete r) throws JessException {
        r.addDeftemplate(getRootTemplate());
        r.addDeftemplate(getNullTemplate());
        r.addDeftemplate(getClearTemplate());
        r.addDeftemplate(getInitialTemplate());
    }

    private boolean m_backchain;
    private final String m_baseName;
    private final String m_fullName;
    private String m_docstring = "";
    private Deftemplate m_parent;
    final ValueVector m_data = new ValueVector();
    private final HashMap m_indexes;
    private final String m_module;

    /**
     * Compare this deftemplate to another one for equality. Two templates are equal if
     * the have the same full name, the same parent, and all the same slots.
     * @param o another template to compare this this one
     * @return true if the templates are the same.
     */
    public boolean equals(Object o) {
        if (!(o instanceof Deftemplate))
            return false;
        Deftemplate t = (Deftemplate) o;
        return (m_fullName.equals(t.m_fullName) &&
                (m_backchain == t.m_backchain) &&
                (m_parent == t.m_parent) &&
                m_data.equals(t.m_data));
    }

    /**
     * Return a hash code for this template.
     * @return the hashcode
     */
    public int hashCode() {
        return m_fullName.hashCode();
    }

    /**
     * Return the parent of this deftemplate. The parent is another
     * deftemplate this one extends. If this template extends no
     * other template explicitly, then it extends the root template.
     * @return the parent deftemplate
     */
    public Deftemplate getParent() {
        return m_parent;
    }

    /**
     * Sever the link with this deftemplate's parent. Useful when
     * creating similar, but unrelated deftemplates.
     */
    private void forgetParent() {
        m_parent = s_rootTemplate;
    }


    /**
     * Get the name of this deftemplate. The name is qualified by the module name.
     * @return the name of this deftemplate
     */
    public final String getName() {
        return m_fullName;
    }

    /**
     * Get the name of this deftemplate unqualified by the module name.
     * @return Tte name of this deftemplate
     */
    public final String getBaseName() {
        return m_baseName;
    }

    /**
     * Get the docstring of this deftemplate.
     * @return The docstring
     */
    public final String getDocstring() {
        return m_docstring;
    }

    /**
     * Make this deftemplate backwards-chaining reactive. Sets a flag in this template, and
     * creates the backward-chaining trigger template.
     * @param engine the engine the new template is created in
     * @throws JessException if anything goes wrong
     */
    public final void doBackwardChaining(Rete engine) throws JessException {
        m_backchain = true;
        Deftemplate newDt = getBackchainingTemplate(engine);
        newDt.forgetParent();
        engine.addDeftemplate(newDt);
    }

    /**
     * Get the backchaining reactivity of this deftemplate.
     * @return True if this deftemplate can stimulate backwards chaining.
     */
    public final boolean getBackwardChaining() {
        return m_backchain;
    }

    /**
     * Create a template that extends the root template.
     * @param name the deftemplate name
     * @param docstring the deftemplate's documentation string
     * @param engine the rule engine to create the template in
     * @throws JessException if anything goes wrong
     */
    public Deftemplate(String name, String docstring, Rete engine)
            throws JessException {

        this(name, docstring, s_rootTemplate, engine);
    }

    /**
     * Create a deftemplate 'derived from' another
     * one. If the name contains a module name, it will be
     * used. Otherwise, the template will be in the current module.
     * @param name the deftemplate name
     * @param docstring the deftemplate's documentation string
     * @param dt the 'parent' of this deftemplate
     * @param engine the engine to create the template in
     * @throws JessException if anything goes wrong
     */

    public Deftemplate(String name, String docstring, Deftemplate dt, Rete engine)
            throws JessException {
        // ###

        int colons = name.indexOf("::");
        if (colons != -1) {
            m_module = name.substring(0, colons);
            engine.verifyModule(m_module);
            m_baseName = name.substring(colons + 2);
            m_fullName = name;
        } else {
            m_module = engine.getCurrentModule();
            m_baseName = name;
            m_fullName = engine.resolveName(name);
        }

        m_parent = dt;
        m_docstring = docstring;

        for (int i = 0; i < dt.m_data.size(); i++)
            m_data.add(dt.m_data.get(i));

        m_indexes = (HashMap) dt.m_indexes.clone();
    }

    /*
     * Used only to construct (not), (test), (__fact), etc.
     */

    private Deftemplate(String name, String docstring) {
        // ###
        m_module = Defmodule.MAIN;
        m_baseName = name;
        m_fullName = RU.scopeName(m_module, m_baseName);
        m_docstring = docstring;
        if (!name.equals(RU.ROOT_DEFTEMPLATE))
            m_parent = s_rootTemplate;
        m_indexes = new HashMap();
    }


    /**
     * Create a new slot in this deftemplate. If the slot already
     * exists, just change the default value. The type parameter is currently stored
     * with the deftemplate, but not used by Jess for any purpose.
     * @param name name of the slot
     * @param value default value for the slot
     * @param typename type of the slot: INTEGER, FLOAT, ANY, etc.
     * @exception JessException If something goes wrong
     */
    public void addSlot(String name, Value value, String typename)
            throws JessException {
        int idx;
        Value type = getSlotTypeCode(typename);

        // Just set default if duplicate
        if ((idx = getSlotIndex(name)) != -1) {
            m_data.set(value, (idx * RU.DT_SLOT_SIZE) + RU.DT_DFLT_DATA);
            m_data.set(type, (idx * RU.DT_SLOT_SIZE) + RU.DT_DATA_TYPE);

        } else {
            int start = m_data.size();
            m_data.setLength(start + RU.DT_SLOT_SIZE);
            m_data.set(new Value(name, RU.SLOT), start + RU.DT_SLOT_NAME);
            m_data.set(value, start + RU.DT_DFLT_DATA);
            m_data.set(type, start + RU.DT_DATA_TYPE);
            m_indexes.put(name, new Integer(start / RU.DT_SLOT_SIZE));
        }
    }

    static boolean isValidSlotType(String typename) {
        return s_typeCodes.containsKey(typename.toUpperCase());
    }

    private static Value getSlotTypeCode(String typename) throws JessException {
        Value type = (Value) s_typeCodes.get(typename.toUpperCase());
        if (type == null)
            throw new JessException("Deftemplate.getSlotTypeCode",
                                    "Bad slot type:", typename);
        return type;
    }

    static String getSlotTypeName(Value typecode) throws JessException {
        String type = (String) s_typeNames.get(typecode);
        if (type == null)
            throw new JessException("Deftemplate.getSlotTypeName",
                                    "Bad slot type:", typecode.toString());
        return type;
    }

    /**
     * Create a new multislot in this deftemplate. If the slot already
     * exists, just change the default value. The type parameter is currently stored
     * with the deftemplate, but not used by Jess for any purpose.
     * @param name name of the slot
     * @param value default value for the slot
     * @param typename name of Jess data type for slot contents
     * @exception JessException if something goes wrong
     */
    public void addMultiSlot(String name, Value value, String typename) throws JessException {
        int idx;

        Value type = getSlotTypeCode(typename);

        // Just set default if duplicate
        if ((idx = getSlotIndex(name)) != -1) {
            m_data.set(value, absoluteIndex(idx) + RU.DT_DFLT_DATA);
        } else {
            int start = m_data.size();

            m_data.setLength(start + RU.DT_SLOT_SIZE);
            m_data.set(new Value(name, RU.MULTISLOT), start + RU.DT_SLOT_NAME);
            m_data.set(value, start + RU.DT_DFLT_DATA);
            m_data.set(type, start + RU.DT_DATA_TYPE);
            m_indexes.put(name, new Integer(start / RU.DT_SLOT_SIZE));
        }
    }

    /**
     * Create a new multislot in this deftemplate. If the slot already
     * exists, just change the default value. The type parameter is currently stored
     * with the deftemplate, but not used by Jess for any purpose. The slot type will
     * be set to "ANY".
     * @param name name of the slot
     * @param value default value for the slot
     * @exception JessException if something goes wrong
     */
    public void addMultiSlot(String name, Value value) throws JessException {
        addMultiSlot(name, value, "ANY");
    }

    private static int absoluteIndex(int index) {
        return index * RU.DT_SLOT_SIZE;
    }

    /**
     * Returns the slot data type (one of the constants in jess.RU)
     * for the slot given by the zero-based index.
     * @param index The zero-based index of the slot (0, 1, 2 ... getNSlots()-1)
     * @return The data type of that slot (RU.INTEGER, RU.SYMBOL, etc., or RU.ANY)
     * @throws JessException if anything goes wrong
     */

    public int getSlotDataType(int index)
            throws JessException {
        return m_data.get(absoluteIndex(index) + RU.DT_DATA_TYPE).intValue(null);
    }

    /**
     * Returns the default value of a slot given by the zero-based
     * index.
     * @param index The zero-based index of the slot (0, 1, 2 ... getNSlots()-1)
     * @return The default value for that slot (can be Funcall.NIL or Funcall.NILLIST for none
     * @throws JessException if anything goes wrong
     */

    public Value getSlotDefault(int index)
            throws JessException {
        return m_data.get(absoluteIndex(index) + RU.DT_DFLT_DATA);
    }

    /**
     * Returns the slot type (RU.SLOT or RU.MULTISLOT) of the slot in
     * this deftemplate given by the zero-based index.
     * @param index The zero-based index of the slot (0, 1, 2 ... getNSlots()-1)
     * @return The type of that slot (RU.SLOT or RU.MULTISLOT)
     * @throws JessException if anything goes wrong
     * */

    public int getSlotType(int index)
            throws JessException {
        return m_data.get(absoluteIndex(index) + RU.DT_SLOT_NAME).type();
    }

    /**
     * Returns the slot type (RU.SLOT or RU.MULTISLOT) of the slot in
     * this deftemplate given by the zero-based index.
     * @param name The name of the slot
     * @return The type of that slot (RU.SLOT or RU.MULTISLOT)
     * @throws JessException if anything goes wrong
     * */

    public int getSlotType(String name)
            throws JessException {
        int index = getSlotIndex(name);
        if (index == -1)
            throw new JessException("Deftemplate.getSlotType",
                                    "No such slot " + name + " in template", m_fullName);
        return getSlotType(index);
    }

    /**
     * Returns true if the slot at the given index is a multislot.
     * @param index the index of the slot
     * @return true if the slot is a multislot
     * @throws JessException if the index is invalid
     */
    public boolean isMultislot(int index) throws JessException {
        return getSlotType(index) == RU.MULTISLOT;
    }

    /**
     * Return the index (0 through getNSlots()-1) of the named slot,
     * or -1 if there is no such slot.
     * @param slotname The name of the slot
     * @return The zero-based index of the slot
     */
    public int getSlotIndex(String slotname) {
        Integer i = (Integer) m_indexes.get(slotname);
        if (i == null)
            return -1;
        else
            return i.intValue();
    }

    /**
     * Return the name of a given slot in this deftemplate.
     * @param index The zero-based index of the slot (0, 1, 2 ... getNSlots()-1)
     * @return The name of that slot
     * @exception JessException If something is horribly wrong
     */
    public String getSlotName(int index) throws JessException {
        return m_data.get(absoluteIndex(index) + RU.DT_SLOT_NAME).
                stringValue(null);
    }

    /**
     * Return a new array containing the names of all the slots in this deftemplate.
     * @return a new array
     */
    public String[] getSlotNames() {
        String[] names = new String[getNSlots()];
        for (int i = 0; i < names.length; i++) {
            try {
                names[i] = getSlotName(i);
            } catch (JessException silentlyIgnore) {
            }
        }
        return names;
    }


    /**
     * Return the number of slots in this deftemplate.
     * @return The number of slots in this deftemplate
     */
    public int getNSlots() {
        return m_data.size() / RU.DT_SLOT_SIZE;
    }

    /**
     * Turn this deftemplate into a String.
     * @return a string representation of the Deftemplate
     */
    public String toString() {
        return "[deftemplate " + m_fullName + "]";
    }

    /**
     * This class participates in the Visitor pattern.
     * @param v a Visitor that is interested in working with this object
     * @return whatever the visitor's visitDeftemplate() method returns                                                         e
     */
    public Object accept(Visitor v) {
        return v.visitDeftemplate(this);
    }

    /**
     * Return the name of the module this deftemplate is in.
     * @return the module name
     */
    public String getModule() {
        return m_module;
    }

    /**
     * Indicate whether this template is a backward chaining trigger.
     * @return true if this template is a backward chaining trigger
     */

    public boolean isBackwardChainingTrigger() {
        return m_baseName.startsWith(RU.BACKCHAIN_PREFIX);
    }

    /**
     * Get the name of the backward chaining trigger template that would be associated with
     * this template, whether it exists or not.
     * @return the template name
     */
    public String getBackchainingTemplateName() {
        return RU.scopeName(m_module, RU.BACKCHAIN_PREFIX + m_baseName);
    }

    /**
     * Return the undecorated template name. If this template is not a backward chaining trigger,
     * then just return its name. If it is, then return the name of the template this template serves as a
     * trigger for.
     * @return the name
     */
    public String getNameWithoutBackchainingPrefix() {
        if (!isBackwardChainingTrigger())
            return m_fullName;
        else
            return RU.scopeName(m_module,
                                m_baseName.substring(RU.BACKCHAIN_PREFIX.length()));
    }


    /**
     * Return a new backward-chaining trigger template for this template.
     * @return the template
     * @param engine the engine containing the template
     * @throws JessException if anything goes wrong
     */

    public Deftemplate getBackchainingTemplate(Rete engine)
            throws JessException {

        return new Deftemplate(getBackchainingTemplateName(),
                               "Goal seeker for " + m_fullName,
                               this, engine);
    }

    /**
     * Return true if this is an ordered template. An ordered template is one that has a single
     * multislot named "__data".
     * @return true if this is an ordered template.
     */
    public boolean isOrdered() {
        return /*getNSlots() == 0 ||*/ getSlotIndex(RU.DEFAULT_SLOT_NAME) == 0;
    }

    /**
     * Set the slot-specific activation behavior for this template.
     * @param b true if this template should use slot-specific behavior
     */
    public void setSlotSpecific(boolean b) {
        m_slotSpecific = b;
    }

    /**
     * Query the slot-specific activation behavior for this template.
     * @return true if this template uses slot-specific behavior
     */
    public boolean isSlotSpecific() {
        return m_slotSpecific;
    }

    /**
     * Return the String "deftemplate".
     * @return "deftemplate"
     */
    public final String getConstructType() {
        return "deftemplate";
    }

    /**
     * If this is a template for shadow facts, return the Java class name.
     * Otherwise, return null.
     * @return a class name, or null
     */
    public String getShadowClassName() {
        if (isShadowTemplate())
            return m_docstring.substring(DefinstanceList.JAVA_OBJECT.length());
        else
            return null;
    }

    /**
     * Return true if this is a template for shadow facts.
     * @return true if this is a shadow template
     */
    public boolean isShadowTemplate() {
        return getSlotIndex("OBJECT") != -1 && m_docstring.indexOf(DefinstanceList.JAVA_OBJECT) == 0;
    }

    /**
     * Return true if this is a template for shadow facts, and public variables were included when it was created.
     * @return true if this is a shadow template with variables included
     * *throws JessException if anything goes wrong
     */
    public boolean includesVariables() throws JessException {
        if (!isShadowTemplate())
            return false;

        Value v = getSlotDefault(getNSlots()-2);

        if (v.type() != RU.JAVA_OBJECT)
            return false;

        return v.javaObjectValue(null) instanceof SerializableVD;
    }


    /**
     * Returns true if the argument is a special template name.
     * @param name the name of a template
     * @return true for "test" or "initial-fact", false otherwise.
     */
    public static boolean isSpecialName(String name) {
        return name.equals("test") || name.equals("initial-fact");
    }

    /**
     * Returns one of the special templates "test" or "initial-fact" by name.
     * @param name "test" or "initial-fact"
     * @return the names template or null
     */
    public static Deftemplate getSpecialTemplate(String name) {
        if (name.equals("test"))
            return s_testTemplate;
        else if (name.equals("initial-fact"))
            return s_initialTemplate;
        else
            return null;
    }

}


