package jess;

import java.io.Serializable;
import java.util.*;

/**
 * Parent class of Defrules and Defqueries.
 * <P>
 * (C) 2006 Sandia National Laboratories<br>
 */

public abstract class HasLHS extends Node
        implements Serializable, Visitable, Modular, NodeSink {
    String m_module;
    String m_name;
    String m_displayName;
    String m_docstring = "";
    private ArrayList m_nodes = new ArrayList();
    private Map m_bindings = new HashMap();
    private Group m_CEs;
    private int m_nodeIndexHash = 0;
    private StringBuffer m_compilationTrace;
    boolean m_new = true;
    boolean m_frozen = false;    
    private HasLHS m_next = null;

    HasLHS(String name, String docstring, Rete engine) throws JessException {

        int colons = name.indexOf("::");
        if (colons != -1) {
            m_module = name.substring(0, colons);
            engine.verifyModule(m_module);
            m_name = name;
        } else {
            m_module = engine.getCurrentModule();
            m_name = engine.resolveName(name);
        }

        int amp = m_name.indexOf('&');
        m_displayName = (amp == -1) ? m_name : m_name.substring(0, amp);

        m_docstring = docstring;
        m_CEs = new Group(Group.AND);
    }

    /**
     * Fetch the number of  elenments on the LHS of this construct.
     * @return The number of CEs
     */
    public int getGroupSize() {
        return m_CEs.getGroupSize();
    }

    /**
     * Return the idx-th Conditional element on this construct's LHS.
     * @param idx The zero-based index of the desired CE
     * @return the CE
     */
    ConditionalElementX getLHSComponent(int idx) {
        return m_CEs.getConditionalElementX(idx);
    }

    /**
     * Consider this ConditionalElement to be READ ONLY!
     */

    public ConditionalElement getConditionalElements() {
        return getLHSComponents();
    }

    ConditionalElementX getLHSComponents() {
        return m_CEs;
    }

    /**
     * Return a string (useful for debugging) describing all the Rete network
     * nodes connected to this construct.
     * @return A textual description of all the nodes used by this construct
     */

    public String listNodes() {
        StringBuffer sb = new StringBuffer(100);
        for (int i = 0; i < m_nodes.size(); i++) {
            sb.append(m_nodes.get(i));
            sb.append("\n");
        }
        return sb.toString();
    }

    /**
     * Some rules that use the "or" conditional element can be compiled into multiple
     * subrules. In those situations, those subrules are chained together and can be
     * accessed, in order, using this method.
     * @return the next subrule or subquery in the chain.
     */
    public HasLHS getNext() {
        return m_next;
    }

    void setNext(HasLHS next) {
        m_next = next;
    }

    void freeze(Rete engine) throws JessException {
        m_frozen = true;
    }

    public int getPatternCount() {
        return getLHSComponents().getPatternCount();
    }

    void insertCEAt(ConditionalElementX ce, int index, Rete engine)
            throws JessException {
        Group CEs = m_CEs;
        m_bindings = new HashMap();
        m_CEs = new Group(Group.AND);
        if (CEs.getGroupSize() == 0) {
            addCE(ce, engine);
            return;
        }

        for (int i = 0; i < CEs.getGroupSize(); i++) {
            if (i == index)
                addCE(ce, engine);
            addCE(CEs.getConditionalElementX(i), engine);
        }
    }

    public void setLHS(Group ce, Rete engine)
            throws JessException {

        ce = (Group) ce.canonicalize();
        createChain(ce, engine);
    }

    abstract void createChain(ConditionalElementX ce, Rete engine)
            throws JessException;

    void doSetLHS(ConditionalElementX ce, Rete engine) throws JessException {
        Group flat = new Group(Group.AND);
        ce.addToGroup(flat);

        for (int i=0; i<ce.getGroupSize(); i++) {
            ConditionalElementX cex = (ConditionalElementX) ce.getConditionalElement(i);
            addCE(cex, engine);
        }
    }

    /**
     * Add a conditional element to this construct
     */

    protected void addCE(ConditionalElementX ce, Rete engine) throws JessException {

        ce = (ConditionalElementX) ce.clone();
        // EJFH This represents an implementation limit, although I'm unsure a limit on *what*
        m_seqNum += 10;
        if (ce.getName().equals(Group.NOT) || ce.getName().equals(Group.ACCUMULATE)) {
            m_CEs.renameVariables(ce, getSequenceNumber());
        }

        storeBoundName(ce);

        m_CEs.add(ce);


        //Look for variables, and create bindings for the ones defined
        //in this CE.
        ce.findVariableDefinitions(getPatternCount() - ce.getPatternCount(), m_bindings, m_bindings);

        // Now look for variables that were never given a proper definition --
        // those that were only negated.
        findUndefinedVariables(ce);

        // Now handle '|' conjunctions by transforming tests into (or) Funcalls
        transformOrConjunctionsIntoOrFuncalls(ce, engine);

        addNOTToSupressUnneededBackwardChaining(ce, engine);

    }

    private void addNOTToSupressUnneededBackwardChaining(ConditionalElementX ce,
                                                         Rete engine)
            throws JessException {
        if (ce.isBackwardChainingTrigger()) {
            Pattern pattern = (Pattern) ce;
            Pattern notPattern =
                    new Pattern(pattern,
                                pattern.getNameWithoutBackchainingPrefix());

            Group not = new Group(Group.NOT);
            not.add(notPattern);
            addCE(not, engine);
        }
    }

    private void storeBoundName(ConditionalElementX ce) {
        String varname = ce.getBoundName();

        if (varname != null && m_bindings.get(varname) == null) {
            if (ce.getName().equals(Group.ACCUMULATE))  {
                addBinding(varname, ce, getGroupSize(), -1, RU.ACCUM_RESULT);
            } else {
                while (! (ce instanceof Pattern))
                    ce = ce.getConditionalElementX(0);
                addBinding(varname, ce, getGroupSize(), -1, RU.PATTERN);
            }
        }
    }

    private void findUndefinedVariables(ConditionalElementX ce)
            throws JessException {
        for (PatternIterator it = new PatternIterator(ce); it.hasNext();) {
            Pattern pattern = (Pattern) it.next();
            for (int i = 0; i < pattern.getNSlots(); i++) {
                for (int j = 0; j < pattern.getNTests(i); j++) {
                    Test1 test = pattern.getTest(i, j);
                    Value val = test.m_slotValue;
                    if (val instanceof Variable) {
                        String name = val.variableValue(null);
                        if (m_bindings.get(name) == null) {
                            // Ignore defglobals here
                            if (Defglobal.isADefglobalName(name))
                                continue;

                            throw new JessException("HasLHS.addPattern",
                                                    "First use of variable negated:",
                                                    RU.removePrefix(name));
                        }

                    } else if (val instanceof FuncallValue) {
                        searchFuncallForUndefinedVariables((FuncallValue) val, (Map) ((HashMap) m_bindings).clone());
                    }
                }
            }
        }
    }

    // CAUTION: Destructively modifies "bindings"
    private void searchFuncallForUndefinedVariables(FuncallValue val, Map bindings) throws JessException {
        Funcall funcall = val.funcallValue(null);
        for (int i=0; i<funcall.size(); ++i) {
            if (funcall.get(i) instanceof Variable) {
                String name = funcall.get(i).variableValue(null);
                if (isADefinition(funcall, i))
                    bindings.put(name, name);
                else if (bindings.get(name) == null && !Defglobal.isADefglobalName(name)) {
                    throw new JessException("HasLHS.addPattern",
                                            "Variable used before definition:",
                                            name);
                }
            } else if (funcall.get(i) instanceof FuncallValue) {
                searchFuncallForUndefinedVariables((FuncallValue) funcall.get(i), bindings);
            }
        }
    }

    private boolean isADefinition(Funcall funcall, int i) throws JessException {
        String name = funcall.get(0).stringValue(null);
        if (i == 1) {
            if (name.equals("bind") || name.equals("foreach"))
                return true;
        }
        return false;
    }

    private void transformOrConjunctionsIntoOrFuncalls(ConditionalElementX ce,
                                                       Rete engine)
            throws JessException {

        PatternIterator it = new PatternIterator(ce);
        for (int index = getPatternCount() - ce.getPatternCount(); it.hasNext(); ++index) {
            Pattern pattern = (Pattern) it.next();
            Deftemplate dt = pattern.getDeftemplate();

            for (int i = 0; i < pattern.getNSlots(); i++) {
                int nTests = pattern.getNTests(i);

                if (nTests == 0)
                    continue;

                // Rearrange the tests
                ArrayList tests = new ArrayList();

                String slotName = dt.getSlotName(i);
                int currentSubIndex = pattern.getTest(i, 0).m_subIdx;
                int doneIdx = 0;

                // This is a loop over sub-indexes in the test array. doneIdx
                // will be incremented inside the loop, and some constructs
                // will break out to this label.
                subIdxLoop:
                    while (doneIdx < nTests) {
                        // Find out if there are any ORs on this subslot
                        boolean hasOrs = false;
                        for (int j = doneIdx; j < nTests; j++) {
                            Test1 aTest = pattern.getTest(i, j);
                            if (aTest.m_subIdx != currentSubIndex)
                                break;
                            else if (aTest.m_conjunction == RU.OR) {
                                hasOrs = true;
                                break;
                            }
                        }

                        // If no ORs on this subslot, just copy tests into ArrayList
                        if (!hasOrs) {
                            Test1 aTest;
                            for (int j = doneIdx; j < nTests; j++) {
                                aTest = pattern.getTest(i, j);
                                if (aTest.m_subIdx != currentSubIndex) {
                                    currentSubIndex = aTest.m_subIdx;
                                    continue subIdxLoop;
                                } else {
                                    tests.add(aTest);
                                    ++doneIdx;
                                }
                            }
                            continue;
                        }


                        // First find a variable to represent this (sub)slot; we
                        // may have to create one
                        Value var;
                        Test1 firstTest = pattern.getTest(i, doneIdx);
                        Value testValue = firstTest.m_slotValue;

                        if (isAVariableDefinition(testValue, pattern, i)) {
                            var = testValue;
                            ++doneIdx;
                        } else {
                            String name = RU.gensym(Tokenizer.BLANK_PREFIX);
                            var = new Variable(name, RU.VARIABLE);
                            m_bindings.put(name,
                                           new BindingValue(name,
                                                            pattern,
                                                            index,
                                                            i, currentSubIndex,
                                                            dt.getSlotDataType(i)));
                        }

                        tests.add(new Test1(TestBase.EQ, slotName, currentSubIndex, var));

                        // We're going to build up this function call
                        Funcall or = new Funcall("or", engine);

                        // Count how many tests until an OR, so we can omit the
                        // AND if not needed
                        while (true) {
                            int andCount = 1;
                            for (int j = doneIdx + 1; j < nTests; j++) {
                                Test1 aTest = pattern.getTest(i, j);
                                if (aTest.m_conjunction == RU.OR ||
                                        aTest.m_subIdx != currentSubIndex)
                                    break;
                                else
                                    ++andCount;
                            }

                            if (andCount == 1) {
                                or.add(testToFuncall(pattern.getTest(i, doneIdx),
                                                     var, engine));
                            } else {
                                Funcall and = new Funcall("and", engine);
                                for (int j = doneIdx; j < doneIdx + andCount; j++)
                                    and.add(testToFuncall(pattern.getTest(i, j),
                                                          var, engine));
                                or.add(new FuncallValue(and));
                            }

                            doneIdx += andCount;

                            if (doneIdx == nTests)
                                break;
                            else if (pattern.getTest(i, doneIdx).m_subIdx !=
                                    currentSubIndex)
                                break;
                        }
                        tests.add(new Test1(TestBase.EQ, slotName, currentSubIndex, new FuncallValue(or)));

                        if (doneIdx < nTests &&
                                pattern.getTest(i, doneIdx).m_subIdx != currentSubIndex)
                            currentSubIndex = pattern.getTest(i, doneIdx).m_subIdx;
                    }

                Test1[] testArray = new Test1[tests.size()];
                for (int j = 0; j < testArray.length; j++)
                    testArray[j] = (Test1) tests.get(j);
                pattern.replaceTests(i, testArray);
            }
        }
    }

    private boolean isAVariableDefinition(Value testValue,
                                          ConditionalElementX pattern,
                                          int slot) throws JessException {

        if (testValue.type() != RU.VARIABLE)
            return false;

        String name = testValue.variableValue(null);
        if (Defglobal.isADefglobalName(name))
            return false;

        BindingValue bv = (BindingValue) m_bindings.get(name);

        return (bv.getCE() == pattern && bv.getSlotIndex() == slot);
    }


    /**
     *  Given a test, create an implied function call that does it
     */
    private Value testToFuncall(Test1 t, Value var, Rete engine)
            throws JessException {
        Value v = t.m_slotValue;
        switch (t.m_slotValue.type()) {
            case RU.FUNCALL:
                {
                    if (t.m_test == TestBase.NEQ)
                        return new FuncallValue(new Funcall("not", engine).arg(v));
                    else
                        return v;
                }
            default:
                return new FuncallValue(new Funcall(t.m_test == TestBase.EQ ?
                        "eq" : "neq", engine).
                        arg(v).arg(var));
        }
    }

    private void addBinding(String name, ConditionalElementX patt,
                            int factidx,
                            int subidx, int type){
        m_bindings.put(name,
                       new BindingValue(name, patt, factidx, type, subidx, RU.FACT));
    }

    Map getBindings() {
        return m_bindings;
    }

    public Iterator getNodes() {
        return m_nodes.iterator();
    }
    public int getNodeCount() {
        return m_nodes.size();
    }


    void ready(Token fact_input, Context c) {

        Fact fact;
        // set up the variable table
        for (Iterator e = getBindings().values().iterator(); e.hasNext();) {
            BindingValue b = (BindingValue) e.next();

            if (b.getSlotIndex() == RU.LOCAL)
            // no default binding for locals
                continue;

            // all other variables need info from a fact
            // if this is a not CE, skip it;
            Value val;
            fact = fact_input.fact(b.getFactNumber());
            try {
                if (b.getSlotIndex() == RU.PATTERN) {
                    val = new FactIDValue(fact.getIcon());
                } else {
                    if (b.getSubIndex() == -1) {
                        val = fact.get(b.getSlotIndex());
                    } else {
                        ValueVector vv = fact.get(b.getSlotIndex()).listValue(c);
                        val = vv.get(b.getSubIndex());
                    }

                }
                c.setVariable(b.getName(), val);
            } catch (Throwable t) {
                // bad binding. These can come from unused bindings in
                // not CE's.
            }
        }
    }

    public void addNode(Node n) throws JessException {
        if (n == null)
            new JessException("HasLHS.addNode",
                              "Compiler fault",
                              "null Node added");
        else {
            for (int i = 0; i < m_nodes.size(); ++i)
                if (n == m_nodes.get(i))
                    return;

            appendCompilationTrace(n);

            ++n.m_usecount;
            m_nodes.add(n);
        }
    }

    /**
     * Completely remove this construct from the Rete network, including
     * removing any internal nodes that are only used by this construct.
     * @param root The roots of the Rete network where this construct lives.
     */

    void remove(Node root) {
        Iterator e = m_nodes.iterator();
        while (e.hasNext()) {
            Node s = (Node) e.next();
            if (--s.m_usecount <= 0) {
                root.removeSuccessor(s);
                Iterator e2 = m_nodes.iterator();
                while (e2.hasNext()) {
                    Node n = (Node) e2.next();
                    n.removeSuccessor(s);
                }
            }
        }
        m_nodes.clear();
        if (getNext() != null)
            getNext().remove(root);
    }

    private void appendCompilationTrace(Node n) {
        if (m_compilationTrace == null)
            m_compilationTrace = new StringBuffer(m_name + ": ");

        if (n.m_usecount == 0)
            m_compilationTrace.append("+");
        else
            m_compilationTrace.append("=");

        m_compilationTrace.append(n.getCompilationTraceToken());
    }

    String getCompilationTraceToken() {
        return "t";
    }

    StringBuffer getCompilationTrace() {
        return m_compilationTrace;
    }

    /**
     * Set the node-index-hash of this construct. The node-index-hash
     * value effects the indexing efficiency of the join nodes for this
     * construct. Larger values will make constructs with many partial
     * matches faster (to a point). Must be set before construct is
     * added to engine (so is typically set during parsing via the
     * equivalent Jess command.
     @param h The node index hash value
     */

    public void setNodeIndexHash(int h) {
        m_nodeIndexHash = h;
        if (getNext() != null)
            getNext().setNodeIndexHash(h);
    }

    /**
     * Get the node-index-hash setting of this construct.
     * @return The node-index-hash of this construct
     */
    public int getNodeIndexHash() {
        return m_nodeIndexHash;
    }

    /**
     * Fetch the name of this construct
     * @return The name of this construct
     */
    public final String getName() {
        return m_name;
    }

    /**
     * Fetch the display name of this construct
     * @return The display name of this construct
     */
    public String getDisplayName() {
        return m_displayName;
    }

    /**
     * Get the documentation string for this construct.
     * @return The docstring for this construct
     */
    public final String getDocstring() {
        return m_docstring;
    }

    // Compiler calls this after we've had initial update
    void setOld() {
        m_new = false;
    }

    public abstract Object accept(Visitor jv);

    public String getModule() {
        return m_module;
    }

    private int m_seqNum;

    int getSequenceNumber() {
        return m_seqNum;
    }

    public int getNodeType() {
        return TYPE_TERMINAL;
    }
}



